from hcai_models.core.abstract_model import Model


class AudioModel(Model):
    """
    Abstract base class for audio models.
    Specifies specific functionality of audio models and ensures compatibility with the interface for external calls.
    """

    def __init__(self, *args, in_sr=16000, n_channels=1, **kwargs):
        kwargs.setdefault("input_shape", (in_sr, n_channels))
        assert len(kwargs["input_shape"]) == 2
        super().__init__(*args, **kwargs)

    def get_input_sample_rate(self):
        return self.input_shape[0]

    def get_n_channels(self):
        return self.input_shape[1]



import tensorflow as tf
import tensorflow.keras as keras
from hcai_models.core.abstract_image_model import ImageModel
from hcai_models.core.keras_model import KerasModel
from tensorflow.python.keras.layers import VersionAwareLayers

layers = VersionAwareLayers()

_DESCRIPTION = """
Convolutional networks are at the core of most state-of-the-art computer vision solutions for a wide variety of tasks. 
Since 2014 very deep convolutional networks started to become mainstream, yielding substantial gains in various benchmarks. 
Although increased model size and computational cost tend to translate to immediate quality gains for most tasks (as long as enough labeled data is provided for training), 
computational efficiency and low parameter count are still enabling factors for various use cases such as mobile vision and big-data scenarios. 
Here we explore ways to scale up networks in ways that aim at utilizing the added computation as efficiently as possible by suitably factorized convolutions and aggressive regularization. 
We benchmark our methods on the ILSVRC 2012 classification challenge validation set demonstrate substantial gains over the state of the art: 
21.2% top-1 and 5.6% top-5 error for single frame evaluation using a network with a computational cost of 5 billion multiply-adds per inference and with using less than 25 million parameters.
With an ensemble of 4 models and multi-crop evaluation, we report 3.5% top-5 error on the validation set (3.6% error on the tests set) and 17.3% top-1 error on the validation set.
"""

_CITATION = """
@inproceedings{szegedy2016rethinking,
  title={Rethinking the inception architecture for computer vision},
  author={Szegedy, Christian and Vanhoucke, Vincent and Ioffe, Sergey and Shlens, Jon and Wojna, Zbigniew},
  booktitle={Proceedings of the IEEE conference on computer vision and pattern recognition},
  pages={2818--2826},
  year={2016}
}
"""


class InceptionV3(ImageModel, KerasModel):
    def __init__(self, *args, **kwargs):
        # Overwriting default parameters
        kwargs.setdefault("input_height", 299)
        kwargs.setdefault("input_width", 299)
        kwargs.setdefault("input_channels", 3)

        # Init parents
        super().__init__(
            *args,
            keras_build_func=keras.applications.inception_v3.InceptionV3,
            **kwargs
        )

    def _build_model(self):
        # Adding preprocessing layer
        inputs = tf.keras.Input((None, None, self.n_channels))
        x = layers.Resizing(self.in_height, self.in_width)(inputs)
        x = layers.Rescaling(1.0 / 127.0, offset=-1)(x)

        # Getting base model
        x = keras.applications.InceptionV3(input_tensor=x, include_top=False, weights=None).output

        # Apply pooling
        if self.pooling == "avg":
            x = layers.GlobalAveragePooling2D()(x)
        elif self.pooling == "max":
            x = layers.GlobalMaxPooling2D()(x)

        # Adding top
        if self.include_top:
            if self.dropout_rate > 0:
                x = layers.Dropout(self.dropout_rate, name="top_dropout")(x)

            if not len(self.output_shape) == 1:
                raise ValueError('Got output shape with multiple dimensions, but only classification / regression tops are supported for InceptionV3')
            x = layers.Dense(
                self.output_shape[0],
                activation=self.output_activation_function,
                name="predictions",
            )(x)

        return keras.Model(inputs=inputs, outputs=x, name=self.name)

    def _info(self):
        return ""

    def preprocess_input(self, ds):
        """
        All preprocessing is done in the layers of the model itself
        :param ds:
        :return:
        """
        # Resizing the input data
        # size = (self.input_shape[0], self.input_shape[1])
        # ds = ds.map(lambda image, label: (tf.image.resize(image, size), label))

        from tensorflow.python.keras.applications.inception_v3 import preprocess_input

        return ds



import numpy as np


class Weights:

    # Overwrite to include a description of the weights.csv in the documentation
    _DESCRIPTION = ""

    def __init__(
        self,
        download_url,
        hash,
        output_shape=None,
        download_url_no_top=None,
        hash_no_top=None,
    ):
        """
        Initialize the weigth datastructure
        :param download_url: The url from where to download the weights.csv
        :param hash: The sha256 hash of the weight file
        :param output_shape: Shape of the outputs of the network. Only necessary if the model is build with the original top
        :param download_url_no_top: Download path for the weightfile with no model top
        :param hash_no_top: The sha256 hash of the weight file with no model top
        """
        self.download_url = download_url
        self.hash = hash
        self.download_url_no_top = download_url_no_top
        self.hash_no_top = hash_no_top
        self.output_shape = (
            (output_shape,) if np.isscalar(output_shape) else output_shape
        )



# Import models
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB0
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB1
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB2
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB3
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB4
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB5
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB6
from hcai_models.models.image.hcai_efficient_net.efficient_net import EfficientNetB7



# Lint as python3
# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for file download and caching. Partially transferred from https://github.com/tensorflow/tensorflow/blob/v2.5.0/tensorflow/python/keras/utils/data_utils.py#L148-L277"""
import hashlib
import os
import shutil
import sys
import tarfile
import urllib
import zipfile

from six.moves.urllib.request import urlopen
from tensorflow.python.keras.utils.generic_utils import Progbar
from tensorflow.python.keras.utils.io_utils import path_to_string

if sys.version_info[0] == 2:

    def urlretrieve(url, filename, reporthook=None, data=None):
        """Replacement for `urlretrieve` for Python 2.

        Under Python 2, `urlretrieve` relies on `FancyURLopener` from legacy
        `urllib` module, known to have issues with proxy management.

        Args:
            url: url to retrieve.
            filename: where to store the retrieved data locally.
            reporthook: a hook function that will be called once on establishment of
              the network connection and once after each block read thereafter. The
              hook will be passed three arguments; a count of blocks transferred so
              far, a block size in bytes, and the total size of the file.
            data: `data` argument passed to `urlopen`.
        """

        def chunk_read(response, chunk_size=8192, reporthook=None):
            content_type = response.info().get("Content-Length")
            total_size = -1
            if content_type is not None:
                total_size = int(content_type.strip())
            count = 0
            while True:
                chunk = response.read(chunk_size)
                count += 1
                if reporthook is not None:
                    reporthook(count, chunk_size, total_size)
                if chunk:
                    yield chunk
                else:
                    break

        response = urlopen(url, data)
        with open(filename, "wb") as fd:
            for chunk in chunk_read(response, reporthook=reporthook):
                fd.write(chunk)


else:
    from urllib.request import urlretrieve  # pylint: disable=g-importing-member


def _extract_archive(file_path, path=".", archive_format="auto"):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    Args:
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    Returns:
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, str):
        archive_format = [archive_format]

    file_path = path_to_string(file_path)
    path = path_to_string(path)

    for archive_type in archive_format:
        if archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(
    fname,
    origin,
    untar=False,
    md5_hash=None,
    file_hash=None,
    cache_subdir="weigths",
    hash_algorithm="auto",
    extract=False,
    archive_format="auto",
    cache_dir=None,
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.hcai_models`, placed in the cache_subdir `weights`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.hcai_models/weights/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    Example:

    ```python
    path_to_downloaded_file = tf.keras.utils.get_file(
        "flower_photos",
        "https://storage.googleapis.com/download.tensorflow.org/example_images/flower_photos.tgz",
        untar=True)
    ```

    Args:
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        untar: Deprecated in favor of `extract` argument.
            boolean, whether the file should be decompressed
        md5_hash: Deprecated in favor of `file_hash` argument.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are `'md5'`, `'sha256'`, and `'auto'`.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are `'auto'`, `'tar'`, `'zip'`, and `None`.
            `'tar'` includes tar, tar.gz, and tar.bz files.
            The default `'auto'` corresponds to `['tar', 'zip']`.
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the default directory `~/.hcai_models/`.

    Returns:
        Path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser("~"), ".hcai_models")

    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = "md5"
    datadir_base = os.path.expanduser(cache_dir)

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join("/tmp", ".hcai_models")
    datadir = os.path.join(datadir_base, cache_subdir)
    _makedirs_exist_ok(datadir)

    fname = path_to_string(fname)

    if untar:
        untar_fpath = os.path.join(datadir, fname)
        fpath = untar_fpath + ".tar.gz"
    else:
        fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print(
                    "A local file was found, but it seems to be "
                    "incomplete or outdated because the "
                    + hash_algorithm
                    + " file hash does not match the original value of "
                    + file_hash
                    + " so we will re-download the data."
                )
                download = True
    else:
        download = True

    if download:
        print("Downloading data from", origin)

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size == -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                urlretrieve(origin, fpath, dl_progress)
            except urllib.error.HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except urllib.error.URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar = None

    if untar:
        if not os.path.exists(untar_fpath):
            _extract_archive(fpath, datadir, archive_format="tar")
        return untar_fpath

    if extract:
        _extract_archive(fpath, datadir, archive_format)

    return fpath


def _makedirs_exist_ok(datadir):
    os.makedirs(datadir, exist_ok=True)  # pylint: disable=unexpected-keyword-arg


def _resolve_hasher(algorithm, file_hash=None):
    """Returns hash algorithm as hashlib function."""
    if algorithm == "sha256":
        return hashlib.sha256()

    if algorithm == "auto" and file_hash is not None and len(file_hash) == 64:
        return hashlib.sha256()

    # This is used only for legacy purposes.
    return hashlib.md5()


def _hash_file(fpath, algorithm="sha256", chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    Example:

    ```python
    _hash_file('/path/to/file.zip')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    Args:
        fpath: path to the file being validated
        algorithm: hash algorithm, one of `'auto'`, `'sha256'`, or `'md5'`.
            The default `'auto'` detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        The file hash
    """
    if isinstance(algorithm, str):
        hasher = _resolve_hasher(algorithm)
    else:
        hasher = algorithm

    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm="auto", chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    Args:
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        Whether the file is valid
    """
    hasher = _resolve_hasher(algorithm, file_hash)

    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False



import csv
import pathlib
import sys
import ast
from abc import ABC, abstractmethod

import hcai_models.utils.data_utils as data_utils
from hcai_models.core.registered_model import RegisteredModel
from hcai_models.core.ssi_compat import SSIModel, SSIBridgeModel
from hcai_models.core.weights import Weights


class Model(ABC, RegisteredModel):
    """
    Abstract base class for all models.
    Specifies general functionality of the models and ensures compatibility with the interface for external calls.
    """

    def __init__(
        self,
        *args,
        model_name=None,
        input_shape=None,
        output_shape=None,
        include_top=False,
        dropout_rate=0.2,
        weights=None,
        output_activation_function="softmax",
        optimizer="adam",
        loss="categorical_crossentropy",
        pooling="avg",
        **kwargs
    ):

        self._available_weights = self._init_weights()
        self.info = self._info()
        self._model = None

        self.model_name = model_name if model_name else self.__class__.__name__
        self.input_shape = input_shape
        self.include_top = include_top
        self.dropout_rate = dropout_rate
        self.weights = weights
        self.output_shape = (
            output_shape if output_shape else self._determine_output_shape()
        )
        self.output_activation_function = output_activation_function
        self.pooling = pooling
        self.optimizer = optimizer
        self.loss = loss

    # Public
    def build_model(self):
        self._model = self._build_model()

    def is_ssi_model(self):
        return issubclass(self.__class__, SSIModel)

    def is_ssi_bridge_model(self):
        return issubclass(self.__class__, SSIBridgeModel)

    # Private
    def _init_weights(self):
        # Loading entries from the weights.csv file from the directory of the model module
        module_path = sys.modules[self.__module__].__file__
        weights_path = pathlib.Path(module_path).parent / "weights.csv"
        weight_dict = {}
        if not weights_path.exists():
            return weight_dict
        with open(weights_path, encoding="UTF-8") as csv_file:
            csv_dict = csv.DictReader(csv_file, delimiter=";")
            list_of_rows = [dict_row for dict_row in csv_dict]

        for row in list_of_rows:
            if row["ModelClass"].lower() == self.__class__.__name__.lower():
                name = row["Name"]
                shape = ast.literal_eval(row["OutputShape"])
                url = row["Download URL"]
                hash = row["Hash"]
                url_no_top = row["Download URL without top"]
                hash_no_top = row["Hash without top"]
                weight_dict[name] = Weights(
                    download_url=url,
                    hash=hash,
                    output_shape=shape,
                    download_url_no_top=url_no_top,
                    hash_no_top=hash_no_top,
                )

        return weight_dict

    def _get_weight_file(self):
        """
        Retrieves the weights for pretrained models. The weights will be loaded from the url specified in the _available_weights for the respective weights of the model.
        Once downloaded the data will be cached at `~/.hcai_models/weights` unless specified otherwise.
        :return: Path to the downloaded weights
        """
        if not self.weights:
            print("No weigths have been specified")
            return None
        if not self.weights in self._available_weights.keys():
            raise ValueError("Specified weights not found in available weights")
        weights = self._available_weights[self.weights]

        if self.include_top:
            file_name = self.model_name + "_" + self.weights + ".h5"
            hash = weights.hash
            url = weights.download_url
        else:
            file_name = self.model_name + "_" + self.weights + "_notop.h5"
            hash = weights.hash_no_top
            url = weights.download_url_no_top
        return data_utils.get_file(
            fname=file_name, origin=url, file_hash=hash, extract=not url.endswith(".h5")
        )

    def _determine_output_shape(self):
        if not self.weights:
            return None
        else:
            if self.weights not in self._available_weights.keys():
                print(
                    "Cannot automatically infer shape.{} not found in available weights.".format(
                        self.weights
                    )
                )
                return None
            else:
                return self._available_weights[self.weights].output_shape

    # Public Abstract
    @abstractmethod
    def preprocess_input(self, ds):
        """
        Preprocesses the input data. Only use for static preprocessing that should be replicated once the model is fully trained (e.g input normalization)
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def compile(self, optimizer, loss, metrics):
        """
        Fixes the models training parameter: loss, metrics and optimizer
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def load_weights(self, filepath):
        """
        Loads weights.csv the weights.csv of a model
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def predict(self, sample):
        """
        Predict the given sample
        Args:
            sample:
        """
        raise NotImplementedError()

    @abstractmethod
    def fit(
        self,
        *args,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        validation_split=0.0,
        validation_data=None,
        **kwargs
    ):
        """
        Fits the model on the provided data. The parameter
        :return:
        """
        raise NotImplemented()

    # Private Abstract Methods
    @abstractmethod
    def _build_model(self):
        """
        Builds the model as specified by the set class parameters and set self.model
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def _info(self):
        """
        Returns additional information for the model
        :return:
        """
        raise NotImplemented()



from abc import ABC, abstractmethod


class SSIModel(ABC):

    # TODO check if we need to make this static
    @abstractmethod
    def get_options(self):
        raise NotImplemented()

    @abstractmethod
    def train(self, data, label_score, opts, vars):
        raise NotImplemented()

    @abstractmethod
    def forward(self, data, probs_or_score, opts, vars):
        raise NotImplemented()

    @abstractmethod
    def load(self, path, opts, vars):
        raise NotImplemented()

    @abstractmethod
    def save(self, path, opts, vars):
        raise NotImplemented()


class SSIBridgeModel(ABC):
    ...



from hcai_models.core.abstract_model import Model


class ImageModel(Model):
    """
    Abstract base class for image models.
    Specifies specific functionality and default values of image models and ensures compatibility with the interface for external calls.
    """

    def __init__(self, *args, input_height=300, input_width=300, input_channels=3, **kwargs):

        self.in_height = input_height
        self.in_width = input_width
        self.n_channels = input_channels

        # Overwriting default parameters
        kwargs.setdefault("input_shape", (input_height, input_width, input_channels))
        assert len(kwargs["input_shape"]) == 3

        # Init parents
        super().__init__(*args, **kwargs)

    def get_input_width(self):
        return self.input_shape[0]

    def get_input_heigth(self):
        return self.input_shape[1]

    def get_n_channels(self):
        return self.input_shape[2]



"""Class to forward the abstract functions required by the abstract_model class to the keras api.
Keras models can inherit the methods implemented by this class to avoid duplicating boilerplate code.
"""

from hcai_models.core.abstract_model import Model


class KerasModel(Model):
    def __init__(self, *args, input_tensor=None, **kwargs):
        self.input_tensor = input_tensor
        super().__init__(*args, **kwargs)

    def load_weights(self, *args, filepath=None, **kwargs):

        filepath = self._get_weight_file()
        self._model.load_weights(*args, filepath=filepath, **kwargs)

    # Pass-through functions to Keras Model API
    def compile(self, *args, **kwargs):
        return self._model.compile(*args, **kwargs)

    def fit(self, *args, **kwargs):
        return self._model.fit(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self._model.predict(*args, **kwargs)



import tensorflow as tf
import tensorflow.keras.applications.efficientnet as keras_efficient_net
import tensorflow.keras as keras

from hcai_models.core.abstract_image_model import ImageModel
from hcai_models.core.keras_model import KerasModel
from tensorflow.python.keras.layers import VersionAwareLayers

layers = VersionAwareLayers()


_DESCRIPTION = """
Convolutional Neural Networks (ConvNets) are commonly developed at a fixed resource budget, and then scaled up for better accuracy if more resources are available. 
In this paper, we systematically study model scaling and identify that carefully balancing network depth, width, and resolution can lead to better performance. 
Based on this observation, we propose a new scaling method that uniformly scales all dimensions of depth/width/resolution using a simple yet highly effective compound coefficient. We demonstrate the effectiveness of this method on scaling up MobileNets and ResNet. 
To go even further, we use neural architecture search to design a new baseline network and scale it up to obtain a family of models, called EfficientNets, which achieve much better accuracy and efficiency than previous ConvNets.
"""

_CITATION = """
@inproceedings{tan2019efficientnet,
  title={Efficientnet: Rethinking model scaling for convolutional neural networks},
  author={Tan, Mingxing and Le, Quoc},
  booktitle={International Conference on Machine Learning},
  pages={6105--6114},
  year={2019},
  organization={PMLR}
}
"""

DENSE_KERNEL_INITIALIZER = {
    "class_name": "VarianceScaling",
    "config": {"scale": 1.0 / 3.0, "mode": "fan_out", "distribution": "uniform"},
}


class EfficientNet(ImageModel, KerasModel):
    def __init__(self, *args, default_input_size, keras_build_func, **kwargs):
        # Overwriting default parameters
        kwargs.setdefault("input_height", default_input_size)
        kwargs.setdefault("input_width", default_input_size)
        kwargs.setdefault("input_channels", 3)

        # Init parents
        super().__init__(*args, **kwargs)

        # Init
        self.keras_build_func = keras_build_func

    def _build_model(self):
        # Adding preprocessing layer
        inputs = tf.keras.Input((None, None, self.n_channels))
        x = layers.Resizing(self.in_height, self.in_width)(inputs)

        # Getting base model
        x = self.keras_build_func(input_tensor=x, include_top=False, weights=None).output

        # Apply pooling
        if self.pooling == "avg":
            x = layers.GlobalAveragePooling2D()(x)
        elif self.pooling == "max":
            x = layers.GlobalMaxPooling2D()(x)

        # Adding top
        if self.include_top:
            if self.dropout_rate > 0:
                x = layers.Dropout(self.dropout_rate, name="top_dropout")(x)

            if not len(self.output_shape) == 1:
                raise ValueError(
                    'Got output shape with multiple dimensions, but only classification / regression tops are supported for InceptionV3')
            x = layers.Dense(
                self.output_shape[0],
                activation=self.output_activation_function,
                name="predictions",
            )(x)

        return keras.Model(inputs=inputs, outputs=x, name=self.name)

    def _info(self):
        return ""

    def preprocess_input(self, ds):
        """
        :param ds:
        :return:
        """

        size = (self.input_shape[0], self.input_shape[1])
        ds = ds.map(lambda image, label: (tf.image.resize(image, size), label))
        return ds


class EfficientNetB0(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=224,
            keras_build_func=keras_efficient_net.EfficientNetB0,
            **kwargs
        )


class EfficientNetB1(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=240,
            keras_build_func=keras_efficient_net.EfficientNetB1,
            **kwargs
        )


class EfficientNetB2(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=260,
            keras_build_func=keras_efficient_net.EfficientNetB2,
            **kwargs
        )


class EfficientNetB3(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=300,
            keras_build_func=keras_efficient_net.EfficientNetB3,
            **kwargs
        )


class EfficientNetB4(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=380,
            keras_build_func=keras_efficient_net.EfficientNetB4,
            **kwargs
        )


class EfficientNetB5(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=456,
            keras_build_func=keras_efficient_net.EfficientNetB5,
            **kwargs
        )


class EfficientNetB6(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=528,
            keras_build_func=keras_efficient_net.EfficientNetB6,
            **kwargs
        )


class EfficientNetB7(EfficientNet):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            default_input_size=600,
            keras_build_func=keras_efficient_net.EfficientNetB7,
            **kwargs
        )



"""Convenience functions to access and mange registered models"""
from hcai_models.core import registered_model as register


def build_model(name, builder_kwargs):
    """
    Builds the model for the given class name if registered.
    If the model has not been registered a module not found exception is raised.
    """

    # If the model is registered as abstract model, not all abstract functions have been implemented.
    if name in register._ABSTRACT_MODEL_REGISTRY.keys():
        try:
            register._ABSTRACT_MODEL_REGISTRY[name]()
        except Exception as ex:
            print("Error instantiating model: {}".format(ex))
            exit()

    if not name in register._MODEL_REGISTRY.keys():
        raise ModuleNotFoundError(
            "No model with name {} has been imported.".format(name)
        )
    model = register._MODEL_REGISTRY[name](**builder_kwargs)
    model.build_model()
    return model



# https://github.com/tensorflow/datasets/blob/2a97ad01378f5e63c410aa3eb36aa688aab62bdd/tensorflow_datasets/core/registered.py#L109
# Internal registry containing <str registered_name, Model subclass>
import inspect

_MODEL_REGISTRY = {}
_ABSTRACT_MODEL_REGISTRY = {}


class RegisteredModel:
    """
    Abstract base class for all weights.csv.
    Registers models upon import.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Set the name if the dataset does not define it.
        # Use __dict__ rather than getattr so subclasses are not affected.
        if not cls.__dict__.get("name"):
            cls.name = cls.__name__

        elif cls.name in _MODEL_REGISTRY:
            raise ValueError(f"Dataset with name {cls.name} already registered.")
        is_abstract = inspect.isabstract(cls)
        if is_abstract:
            _ABSTRACT_MODEL_REGISTRY[cls.name] = cls
        else:
            _MODEL_REGISTRY[cls.name] = cls

    def list_available_models(self):
        all_models = [model_name for model_name, model_class in _MODEL_REGISTRY.items()]
        return sorted(all_models)
